@@ -2,16 +2,17 @@ import asyncio
 import logging
 import string
 import unicodedata
+from math import ceil
 from typing import List, Optional, Union
 
 import audible
 import httpx
 from audible.aescipher import decrypt_voucher_from_licenserequest
-
+from audible.client import convert_response_content
 
 from .constants import CODEC_HIGH_QUALITY, CODEC_NORMAL_QUALITY
-from .exceptions import AudibleCliException
-from .utils import LongestSubString
+from .exceptions import AudibleCliException, NotDownloadableAsAAX
+from .utils import full_response_callback, LongestSubString
 
 
 logger = logging.getLogger("audible_cli.models")
@@ -72,6 +73,27 @@ class BaseItem:
 
         return slug_title
 
+    def create_base_filename(self, mode: str):
+        supported_modes = ("ascii", "asin_ascii", "unicode", "asin_unicode")
+        if mode not in supported_modes:
+            raise AudibleCliException(
+                f"Unsupported mode {mode} for name creation"
+            )
+
+        if "ascii" in mode:
+            base_filename = self.full_title_slugify
+
+        elif "unicode" in mode:
+            base_filename = unicodedata.normalize("NFKD", self.full_title)
+
+        else:
+            base_filename = self.asin
+
+        if "asin" in mode:
+            base_filename = self.asin + "_" + base_filename
+
+        return base_filename
+
     def substring_in_title_accuracy(self, substring):
         match = LongestSubString(substring, self.full_title)
         return round(match.percentage, 2)
@@ -153,7 +175,7 @@ class LibraryItem(BaseItem):
         """
 
         # Only items with content_delivery_type 
-        # MultiPartBook or Periodical have child elemts
+        # MultiPartBook or Periodical have child elements
         if not self.has_children:
             return
 
@@ -189,23 +211,22 @@ class LibraryItem(BaseItem):
     def is_downloadable(self):
         # customer_rights must be in response_groups
         if self.customer_rights is not None:
-            if not self.customer_rights["is_consumable_offline"]:
-                return False
-            else:
+            if self.customer_rights["is_consumable_offline"]:
                 return True
+            return False
 
     async def get_aax_url_old(self, quality: str = "high"):
         if not self.is_downloadable():
             raise AudibleCliException(
-                f"{self.full_title} is not downloadable. Skip item."
+                f"{self.full_title} is not downloadable."
             )
 
         codec, codec_name = self._get_codec(quality)
-        if codec is None:
-            raise AudibleCliException(
+        if codec is None or self.is_ayce:
+            raise NotDownloadableAsAAX(
                 f"{self.full_title} is not downloadable in AAX format"
             )
-        
+
         url = (
             "https://cde-ta-g7g.amazon.com/FionaCDEServiceEngine/"
             "FSDownloadContent"
@@ -238,8 +259,8 @@ class LibraryItem(BaseItem):
             )
 
         codec, codec_name = self._get_codec(quality)
-        if codec is None:
-            raise AudibleCliException(
+        if codec is None or self.is_ayce:
+            raise NotDownloadableAsAAX(
                 f"{self.full_title} is not downloadable in AAX format"
             )
 
@@ -252,6 +273,11 @@ class LibraryItem(BaseItem):
         return httpx.URL(url, params=params), codec_name
 
     async def get_aaxc_url(self, quality: str = "high"):
+        if not self.is_downloadable():
+            raise AudibleCliException(
+                f"{self.full_title} is not downloadable."
+            )
+
         assert quality in ("best", "high", "normal",)
 
         body = {
@@ -292,6 +318,17 @@ class LibraryItem(BaseItem):
 
         return metadata
 
+    async def get_annotations(self):
+        url = f"https://cde-ta-g7g.amazon.com/FionaCDEServiceEngine/sidecar"
+        params = {
+            "type": "AUDI",
+            "key": self.asin
+        }
+
+        annotations = await self._client.get(url, params=params)
+
+        return annotations
+
 
 class WishlistItem(BaseItem):
     pass
@@ -315,9 +352,13 @@ class BaseList:
     def _prepare_data(self, data: Union[dict, list]) -> Union[dict, list]:
         return data
 
+    @property
+    def data(self):
+        return self._data
+
     def get_item_by_asin(self, asin):
         try:
-            return next(i for i in self._data if asin in i.asin)
+            return next(i for i in self._data if asin == i.asin)
         except StopIteration:
             return None
 
@@ -354,6 +395,7 @@ class Library(BaseList):
     async def from_api(
             cls,
             api_client: audible.AsyncClient,
+            include_total_count_header: bool = False,
             **request_params
     ):
         if "response_groups" not in request_params:
@@ -369,8 +411,18 @@ class Library(BaseList):
                 "periodicals, provided_review, product_details"
             )
 
-        resp = await api_client.get("library", **request_params)
-        return cls(resp, api_client=api_client)
+        resp: httpx.Response = await api_client.get(
+            "library",
+            response_callback=full_response_callback,
+            **request_params
+        )
+        resp_content = convert_response_content(resp)
+        total_count_header = resp.headers.get("total-count")
+        cls_instance = cls(resp_content, api_client=api_client)
+
+        if include_total_count_header:
+            return cls_instance, total_count_header
+        return cls_instance
 
     @classmethod
     async def from_api_full_sync(
@@ -379,33 +431,42 @@ class Library(BaseList):
             bunch_size: int = 1000,
             **request_params
     ) -> "Library":
-        request_params["page"] = 1
+        request_params.pop("page", None)
         request_params["num_results"] = bunch_size
 
-        library = []
-        while True:
-            resp = await cls.from_api(api_client, params=request_params)
-            items = resp._data
-            len_items = len(items)
-            library.extend(items)
-            if len_items < bunch_size:
-                break
-            request_params["page"] += 1
+        library, total_count = await cls.from_api(
+            api_client,
+            page=1,
+            params=request_params,
+            include_total_count_header=True,
+        )
+        pages = ceil(int(total_count) / bunch_size)
+        if pages == 1:
+            return library
 
-        resp._data = library
-        return resp
+        additional_pages = []
+        for page in range(2, pages+1):
+            additional_pages.append(
+                cls.from_api(
+                    api_client,
+                    page=page,
+                    params=request_params,
+                )
+            )
+
+        additional_pages = await asyncio.gather(*additional_pages)
+
+        for p in additional_pages:
+            library.data.extend(p.data)
+
+        return library
 
     async def resolve_podcats(self):
-        podcasts = []
-        for i in self:
-            if i.is_parent_podcast():
-                podcasts.append(i)
-
         podcast_items = await asyncio.gather(
-            *[i.get_child_items() for i in podcasts]
+            *[i.get_child_items() for i in self if i.is_parent_podcast()]
         )
         for i in podcast_items:
-            self._data.extend(i._data)
+            self.data.extend(i.data)
 
 
 class Catalog(BaseList):
@@ -465,16 +526,11 @@ class Catalog(BaseList):
         return cls(resp, api_client=api_client)
 
     async def resolve_podcats(self):
-        podcasts = []
-        for i in self:
-            if i.is_parent_podcast():
-                podcasts.append(i)
-
         podcast_items = await asyncio.gather(
-            *[i.get_child_items() for i in podcasts]
+            *[i.get_child_items() for i in self if i.is_parent_podcast()]
         )
         for i in podcast_items:
-            self._data.extend(i._data)
+            self.data.extend(i.data)
 
 
 class Wishlist(BaseList):
